//
// MB408 - Screen reader per Linux
// -------------------------------
//
// Autore:   MDV Ingegneria Elettronica S.A.S.
// Versione: 0.2 del 06/09/99
//
// Copyright (C) 1999 by MDV Ingegneria Elettronica S.A.S., 
// All rights reserved.
//
// MB408 comes with ABSOLUTELY NO WARRANTY.
// 
// This is free software, placed under the terms of the
// GNU General Public License, as published by the Free Software
// Foundation.  Please see the file COPYING for details.
// 
// This software is maintained by MDV S.A.S. <mdvsas@tin.it>.


#include <stdio.h>
#include <termios.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <sys/mman.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/resource.h>


#define bool char
#define byte unsigned char
#define word unsigned short

// Codici tasto (KeyCodes) inviati da MB408 

#define   F1          1
#define   F2          2
#define   F3          3
#define   F4          4
#define   F5          5
#define   F6          6
#define   F7          7
#define   F8          8
#define   F9          9
#define   F0          10
#define   LF          11
#define   UP          12
#define   RG          13
#define   DN          14
#define   S_F1        17
#define   S_F2        18
#define   S_F3        19
#define   S_F4        20
#define   S_F5        21
#define   S_F6        22
#define   S_F7        23
#define   S_F8        24
#define   S_F9        25
#define   S_F0        26
#define   S_LF        27
#define   S_UP        28
#define   S_RG        29
#define   S_DN        30
#define   L_F1        33
#define   L_F2        34
#define   L_F3        35
#define   L_F4        36
#define   L_F5        37
#define   L_F6        38
#define   L_F7        39
#define   L_F8        40
#define   L_F9        41
#define   L_F0        42
#define   SL_F1       49
#define   SL_F2       50
#define   SL_F3       51
#define   SL_F4       52
#define   SL_F5       53
#define   SL_F6       54
#define   SL_F7       55
#define   SL_F8       56
#define   SL_F9       57
#define   SL_F0       58
#define   SL_LF       59
#define   SL_UP       60
#define   SL_RG       61
#define   SL_DN       62
#define   ST_UP       128
#define   N_LF        129
#define   N_RG        130
#define   CUR_ROUTE   63
#define   MOUSE       64

#define up_v 0
#define left_h 1
#define right_h 2
#define down_v 3
#define filter_on  0xAF          // Per il filtro dei caratteri semigrafici }
#define filter_off 0xFE

// Costanti per il curosor routing
#define RIGHTAR   "\x1B[C" // Sequenza di escape della freccia Dx sulla tastiera 
#define LEFTAR    "\x1B[D" // Sequenza di escape della freccia Sx sulla tastiera 
#define UPAR      "\x1B[A" // Sequenza di escape della freccia Su sulla tastiera 
#define DNAR      "\x1B[B" // Sequenza di escape della freccia Gi sulla tastiera 
#define PGUP      "\x1B[5~"// Sequenza di escape pagina S sulla tastiera 
#define PGDN      "\x1B[6~"// Sequenza di escape pagina Gi sulla tastiera 
#define MAXROUTERETRY 10;   // Numero di tentativi max per il cursor routing 

#define START 0
#define RUN 1

#define TRUE 1
#define FALSE 0
#define SOH 0x01
#define STX 0x02
#define ETX 0x03
#define LOOPTIME 30     // Tempo di loop in msec
#define RXTIMEOUT 500   // Timeout ricezione caratteri in msec

#define VCSADEV "/dev/vcsa0"
#define CONSDEV "/dev/console"

#define TITLE "---------------------------------------------------------------------------\n" \
              "MB408 - Terminale Braille a 40 caratteri per PC-IBM e compatibili\n" \
              "---------------------------------------------------------------------------\n" \
              "Progettato e realizzato da: MDV Ingegneria Elettronica SAS - S.Lazzaro (BO)\n" \
              "Distribuito da ...........: Istituto dei Ciechi Francesco Cavazza - Bologna\n" \
              "                            U.I.C. Centro Nazionale Tiflotecnico  - Roma\n" \
              "---------------------------------------------------------------------------\n" \
              "Programma : mb408 - screen reader per Linux\n" \
              "Versione  : 0.3 del 08/09/99\n"

#define send_ack() send_packet(127,0,NULL)
static byte latin1_to_ibm437[128] =
{
   199, 252, 233, 226, 228, 224, 229, 231, 234, 235, 232, 239, 238, 236, 196, 197,
   201, 230, 198, 244, 246, 242, 251, 249, 255, 214, 220, 162, 163, 165, 158, 159, 
   225, 237, 243, 156, 241, 209, 170, 186, 191, 169, 172, 189, 188, 161, 171, 187, 
   248, 164, 253, 179, 180, 145,  20, 156, 184, 185,  21, 175, 166, 174, 190, 168, 
   192, 193, 194, 192, 142, 179, 218, 128, 200, 217, 196, 203, 191, 205, 206, 207, 
   208, 157, 210, 211, 212, 213, 153, 215, 216, 217, 218, 219, 154, 221, 222, 152, 
   133, 223, 131, 227, 132, 134, 181, 135, 138, 130, 136, 137, 141, 173, 140, 139,
   240, 177, 149, 155, 147, 245, 247, 148, 176, 151, 183, 150, 129, 178, 254, 160
};
/*
Sono state modificate le seguenti entry per rendere compatibile la tabella con 
i caratteri semigrafici utilizzati dalle "ncurses" e in particolare le cornici
del midnight commander

0xCA = ascii 196 cornice orizzontale
0xC6 = ascii 218 angolo superiore sinistro
0xCC = ascii 191 angolo superiore destro
0xC5 = ascii 179 cornice verticale
0xC3 = ascii 192 angolo inferiore sinistro
0xC9 = ascii 217 angolo inferiore destro

*/

static const char* com_str[] = {"COM1","COM2","COM3","COM4"};
static const char* dev_str[] = {"/dev/ttyS0","/dev/ttyS1","/dev/ttyS2","/dev/ttyS3"};
static const int buad_rate[2] = {B19200,B38400};
int baud_vel = 0;

struct packet {
   byte code;
   byte len;
   byte data[256];
};

struct scrbuf {
   byte nrows;
   byte ncols;
   byte curx;
   byte cury;
   byte data[8000];
};

char pidstr[16];
int what_com = 0;
struct termios oldtio,newtio;
int comfd = 0;
int vcfd = 0;
int log = 0;
byte cntlog=0; char mybuf[256];
bool port_ok = FALSE;
struct packet pkt;
struct packet last_pkt;  // Memorizza l'ultimo pkt inviato
bool pkt_avl = FALSE;
bool awt_ack = FALSE;
bool ok = FALSE;
char old_x,old_y;
bool found;
char k;
struct scrbuf scr;
bool keep_going = TRUE;
struct sched_param sp;
   

byte WR,WC,WL,WA;    // per la gestione delle finestre

byte time_out_ack;   // Timer per il timeout dei pacchetti di ACK 
bool pkt2send[6];    // Marker per i pacchetti da inviare 
byte t_sync;         // Timer per la sincronizzazione utilizzato con ProgStatus=Start 
byte tx_state;       // Indica lo stato della trasmissione
byte hc_x, hc_y;     // Coordinate il cursore hardware nella gestion del per il soft-cursor
byte video_buf[80];  // Buffer area video puntata 
byte cur_x;          // Posizione X cursore 
byte cur_y;          // Posizione Y cursore 
char mb_x;           // Posizione X MB408 
char mb_y;           // Posizione Y MB408 
byte n_cons;         // Numero della console attiva
byte n_col;          // N.colonne modo video 
byte n_lines;        // Indica il numero di righe del modo video 
byte video_buf_len;  // Lunghezza area video puntata 
byte spare;
byte control_buf[9]; // Buffer di controllo 
bool n_col_changed;  // Indica se c' stata una commutazione 80-40 col.
bool n_line_changed; // Indica se c'e stata una variazione delle righe video.
bool full_screen;    // Indica se  attivo il full screen 
char x_min,x_max,    // Estremi della window video attiva 
     y_min,y_max,
     x_min0,x_max0,  // Estremi della window video definita 
     y_min0,y_max0;
char x_step,mb_ofs,  // Per la gestione offset orizzontale 
     left_step,
     ofs_step,
     right_step;
bool tracking;       // Aggancio cursore si/no (true/false) 
bool trk_type40;     // Tipo di aggancio/avanz. 40/36 car. (true/false)
byte trk_delay;      // Ritardo di aggancio cursore 
bool re_track;       // Per il riaggancio momentaneo cursore 
byte cc;             // Cont. per la stabilizzazione di ControlChanged 
bool cc_flag;
bool prog_status;    // Flag di stato del programma 
byte key_code;       // Codice tasto inviato da MB-408 
char ad_x, ad_y;     // Per l'auto-detect delle finestre 

                    // Per la gestione del Soft-Cursor 
byte sc_atrb[3],sc_mask[3];
byte sc_buffer[4000];
word sc_pos[3][128];
word sc_buflen;
byte sc_posidx[3];
word sc_start,sc_end;
bool sc_tracking;
bool hc_tracking[3];
bool sc_first[3];
bool sc_act[3];
byte sc_row1[3],sc_row2[3];
byte sc_x[3],sc_y[3];
bool oc_x,oc_y,sc_scanact;
byte scnum;
bool mov_ok;         // Utilizzata per lo spostamento della finestra di lettura 
word start_adrs;     // Offset iniziale di MB408 del buffer della memoria video
bool auto_read;      // Indica se  attiva la modalit per leggere un testo senza
                     // dover premere i tasti sulla tastiera del pc 
byte type_auto_read; // 0=lettura di un testo in un editor 
                     //   1=lettura di un testo in un viewer 
bool old_tracking;   // Memorizza lo stato di aggancio cursore 
// Gestione cursor routing 
int cur_routex;      // Colonna che deve raggiungere il cursore 
int cur_routey;      // Riga che deve raggiungere il cursore 
int arr_routex;      // Indica il numero di tasti freccia orinzontali 
int arr_routey;      // Indica il numero di tasti freccia verticali 
byte route_cur;      // Indica 0 se  disattivo il cursor routing 
                     // Indica 1 se  attivo il cursor routing senza aggancio cursore
                     // Indica 2 se  attivo il cursor routing con aggancio cursore 
bool y_ok;           // Indica se ha raggiunto la riga indicata da cur_routey 
byte route_retry;    // Contatore dei tentativi per raggiunge la riga corretta 
byte route_state;    // Indica lo stato di avanzamento del cursor routing 
bool fast_cur;       // Indica se  possibile inserire tutti i tasti nel buffer 
byte next_pos;       // Indica la prossima posizione da raggiungere 


void beep(int freq, int msec)
//
// Esegue un beep alla frequenza indicata per msec millisecondi.
//
{
   struct vt_stat vtstat;
   int fd=open(CONSDEV,O_WRONLY);
   if (fd<0) return;
   ioctl(fd,KIOCSOUND,1190000/freq);
   usleep(msec*1000);
   ioctl(fd,KIOCSOUND,0);
   close(fd);
}

int open_serial()
//
// Apre la porta seriale, riportando -1 in caso di errore.
//
{
   // Apriamo il device in lettura e scrittura e non come controlling tty,
   // per evitare che il programma venga terminato da un CTRL-C eventualmente
   // ricevuto per errore...
   comfd = open(dev_str[what_com], O_RDWR | O_NOCTTY);
   if (comfd<0) return (comfd);

   // Salviamo il setup attuale della porta seriale
   tcgetattr(comfd,&oldtio);

   // Impostiamo il nuovo setup
   bzero(&newtio,sizeof(newtio));

   // CS8: 8 bit, no parity, 1 stop
   // CLOCAL: connessione locale, nessun controllo del modem
   // CREAD: abilita la ricezione
   newtio.c_cflag =  buad_rate[baud_vel] | CS8 | CLOCAL | CREAD;

   // Raw input
   newtio.c_iflag = 0;

   // Raw output
   newtio.c_oflag = 0;

   // non-canonical input processing mode
   newtio.c_lflag = 0;

   // Imposta i parametri di attesa caratteri per la read()
   newtio.c_cc[VMIN]=0;
   newtio.c_cc[VTIME]=0;

   // azzera le queue di lettura/scrittura e attiva il nuovo settaggio
   tcflush(comfd,TCIFLUSH);
   tcsetattr(comfd,TCSANOW,&newtio);

   return(0);
}

void close_serial()
//
// Chiude la porta seriale.
//
{
   // Ripristina il settaggio originale della porta seriale
   tcsetattr(comfd,TCSANOW,&oldtio);
   close(comfd);
}

int program_loaded()
//
// Controlla se il programma  gi caricato.
// Riporta: -1=errore nel controllo, 0=non caricato, >0=caricato.
//
{
   FILE* output;
   char buf[256];
   int i=0,j=0;

   output=popen("pstree -p","r");
   if (!output) return -1;
   while (fgets(buf,256,output)) {
      if (strstr(buf,"|-mb408")) {
         while ((i<255)) {
            if (buf[i]!='(') {
               i++;
            } else {
               while ((i<255)) {
                  if (buf[i]!=')') {
                     pidstr[j++]=buf[i];
                  } else {
                     pidstr[j++]=buf[i];
                     pidstr[j]=0;
                     pclose(output);
                     return 1;
                  }
                  i++;
               }
            }
         }
         pclose(output);
         return -1;
      }
   }
   pclose(output);
   return 0;
}

int send_packet(byte code, byte len, byte* data)
//
// Trasmette un pacchetto con codice, dati e lunghezza indicati.
// Riporta -1 in caso di errore.
//
{
   static byte header[5] = { SOH,STX,0,0,ETX };
   word cks=0;
   int i;

   header[2]=code;
   header[3]=len;
   if (write(comfd,header,5)!=5) return -1;
   for (i=1;i<5;i++) cks+=header[i];
   if (write(comfd,data,len)!=len) return -1;
   while (len--) cks+=*data++;
   cks^=0xAA55;
   if (code!=127) awt_ack=TRUE;
   if (write(comfd,&cks,2)!=2) return -1;
   return 0;
}

void rx_pkt_handler()
//
// Macchina  a stati di gestione pacchetti ricevuti.
//
{
   static struct packet temp_pkt;
   static int state = 0;
   static int rxlen,rxind;
   static word cks;
   static int timeout;
   byte buf[256];
   int ndata,i;
   byte *p;

   ndata=read(comfd,buf,256);
   if (ndata<1) {
      if (state>0 && timeout>0)
         if (--timeout==0) state=0;
      return;
   }

   for (i=0,p=buf;i<ndata;i++,p++) {
      switch (state) {
         case 0: // SOH
            if (*p==SOH) { timeout=RXTIMEOUT/LOOPTIME; state++; }
            break;
         case 1: // STX
            if (*p==STX) { state++; cks=STX; } else state=0;
            break;
         case 2: // code
            temp_pkt.code=*p;
            cks += *p;
            state++;
            break;
         case 3: // len
            temp_pkt.len=rxlen=*p;
            rxind=0;
            cks += *p;
            state++;
            break;
         case 4: // ETX
            if (*p==ETX) { state++; cks+=ETX;} else state=0;
            break;
         case 5: // data/cks
            if (rxind<rxlen) {
               temp_pkt.data[rxind++]=*p;
               cks += *p;
            } else {
               cks ^= 0xAA55;
               cks ^= *p;
               state++;
            }
            break;
         case 6: // cks
            cks ^= (word)(*p)<<8;
            if (cks==0) {
               if (temp_pkt.code==127) awt_ack=FALSE;
               else {
                  memcpy(&pkt,&temp_pkt,rxlen+2);
                  pkt_avl=TRUE;
               }
            }
            state=0;
      }
   }
}

void parse_options(int argc,char *argv[])
//
// Analizza le opzioni specificate per il programma e imposta flag e variabili
// di conseguenza.
//
{
   int c;
   opterr=0;
   while ((c=getopt(argc,argv,":c:b:"))!=-1)
      switch (c) {
         case 'c': 
            what_com=atoi(optarg); 
            if (what_com<1 || what_com>4) {
               printf("Porta seriale non valida.\n");
               exit(1);
            }
            what_com--;
            break;
         case 'b': 
            baud_vel=atoi(optarg); 
            if ((baud_vel!=19200) && (baud_vel!=38400)) {
               printf("Baudrate non valido.\n");
               exit(1);
            }
            baud_vel = baud_vel / 19200;
            break;
         case '?': if (isprint(optopt))
                      printf("Opzione sconosciuta: '-%c'.\n",optopt);
                   else
                      printf("Carattere di opzione sconosciuto: '\\x%x'.\n",optopt);
                   exit(1);
         case ':': printf("Argomento mancante in '-%c'.\n",optopt);
                   exit(1);
         default : printf("Errore sulla linea comandi.\n");
                   exit(1);
      }
}

void close_program()
//
// Operazioni da eseguire in chiusura del programma
//
{
   if (port_ok) close_serial();
   if (vcfd>0) close(vcfd);
   if (log>0) close(log);
}

void SIGTERM_handler(int s)
//
// Gestore segnale SIGTERM.
//
{
   keep_going=FALSE;
   signal(SIGTERM,SIGTERM_handler);
}

void SIGKILL_handler(int s)
//
// Gestore segnale SIGKILL.
//
{
   beep(1000,300);
   write(log,"fine",5);
}

void read_screen()
//
// Legge lo schermo dalla console virtuale
//
{
   int len,i;
   lseek(vcfd,0,SEEK_SET);
   read(vcfd,&scr,4);
   len=(int)scr.nrows*(int)scr.ncols*2;
   if (len>8000) len=8000;
   read(vcfd,&scr.data,len);
   for (i=0;i<len;i+=2) {
      if (scr.data[i]>127) scr.data[i]=latin1_to_ibm437[scr.data[i] & 0x7F];
   }
}

int get_active_console()
//
// Riposta il numero di console attiva, -1 in caso di errore.
//
{
   struct vt_stat vtstat;
   int fd=open(CONSDEV,O_RDONLY);
   if (fd<0) return -1;
   ioctl(fd,VT_GETSTATE,&vtstat);
   close(fd);
   return vtstat.v_active;
}

void upd_control_buf()
//
// Aggiorna il valore di control_buf con le varie variabile di controllo
//
{
   control_buf[0] = cur_x;   
   control_buf[1] = cur_y;   
   control_buf[2] = mb_x;    
   control_buf[3] = mb_y;    
   control_buf[4] = n_cons;  
   control_buf[5] = n_col;   
   control_buf[6] = n_lines;  
   control_buf[7] = video_buf_len; 
}

void reset_prog_state()
// 
// Impostazione parametri per il reset dello stato del programma.
//
{
   int i;
   for (i=0;i<sizeof(pkt2send);i++) pkt2send[i]=0;
   t_sync = 0;
   tx_state = 0;
   hc_x = 0xFF;
   hc_y = 0xFF;
   awt_ack = FALSE;
   for (i=0;i<80;i++) video_buf[i]=0; 
   cur_x=0;   
   cur_y=0;   
   mb_x=0;    
   mb_y=0;    
   n_cons=0;  
   n_col=0;   
   n_lines=0;  
   video_buf_len=0; 
   prog_status=START;
}

bool get_nlines()
//
// Controlla il numero di linee del modo video attivo e imposta la 
// variabile n_lines.
// Riporta FALSE= non  cambiato il modo video.
//         TRUE =  cambiato il modo video.
{
   read_screen(); 
   if (n_lines!=scr.nrows) {
      n_lines=scr.nrows;
      n_line_changed=TRUE;
      return TRUE;
   } else return FALSE;
}

void get_side_steps()
// 
// Calcola il valore di LeftStep e RightStep.
//
{

   char i; 

   i = mb_ofs;
   while (i>=0) i -= x_step;
   left_step = i+x_step;
   ofs_step = i+x_step;
   i = mb_ofs;
   while (i<=(x_max-x_min-39)) i += x_step;
   right_step = x_max-39-x_min-i+x_step;
}

void init()
//
// Inizializzazione generale
//
{
   int i,w;

   reset_prog_state();
   n_lines=255;
   get_nlines();
   x_min = 0;
   x_max = scr.ncols-1;
   y_min = 0;
   y_max = n_lines-1;
   full_screen = TRUE;
   x_min0 = y_min;
   x_min0 = y_min;
   x_max0 = x_max;
   x_max0 = y_max;
   x_step = 40;
   mb_x = 0;
   mb_y = 0;
   mb_ofs = 0;
   video_buf_len = 40;
   get_side_steps();
   cc = 0;
   cc_flag = FALSE;
   tracking = FALSE;
   trk_type40 = TRUE;
   trk_delay = 0;
   n_col_changed = FALSE;
   n_line_changed = FALSE;
   ad_x = 0;
   ad_y = 0;
   key_code = 0;
   for (i=0;i<4000;i++) sc_buffer[i]=0;
   sc_buflen=scr.ncols*scr.nrows;
   for (i=0;i<3;i++) { sc_posidx[i]=sc_x[i]=sc_y[i]=0; }
   sc_tracking = FALSE;
   for (w=0;w<3;w++) {
      for (i=0;i<127;i++) sc_pos[w][i]=0;
      sc_atrb[w] = 0x70;
      sc_mask[w] = 0x70;
      hc_tracking[w] = FALSE;
      sc_first[w] = TRUE;
   };
   oc_x = 0;
   oc_y = 0;
   hc_x = 0;
   hc_y = 0;
   sc_scanact = FALSE;
}


void stuf_key (char* key)
//
// Inserisce nel buffer di tastiera la stringa passata
//
{
   int kbdbuf_fd;			
   long kbmode;

   kbdbuf_fd = open (CONSDEV,O_RDONLY);
   if (kbdbuf_fd == -1) { route_cur=0; return; }
   ioctl(kbdbuf_fd,KDGKBMODE,&kbmode);
   if( kbmode != K_XLATE ) { route_cur=0; return; }
   
   while (*key)
      if (ioctl (kbdbuf_fd,TIOCSTI,key++)) break;
   close (kbdbuf_fd);
}

void stuf_key_str (byte count,char* s)
//
// Inserisce nel buffer di tastiera count volte la stringa "s" 
//
{
   byte b;
   for (b=0;b<count;b++) stuf_key(s);
}

/*****************************************************
FUNZIONI PER LA GESTIONE COMUNICAZIONE CON MB408S *** 
*****************************************************/

void set_pkt (byte cpkt)
//
// Attiva il flag corrispondente nell'array PktToSend[].
// pkt2send[2]  un caso speciale, infatti ne vengono utilizzati 2 bit nei 
// seguenti casi:
//    code=134 : imposta il bit 0
//    code=135 : imposta il bit 1
//    code=136 : imposta i bit 0 e 1
//
// I pacchetti associati ai vari flag sono: 
//    Pkt2Send[0] -> code=130
//    Pkt2Send[1] -> code=132
//    Pkt2Send[2] -> code=134,135,136
//    Pkt2Send[3] -> code=142
//    Pkt2Send[4] -> code=150
//    Pkt2Send[5] -> code=153

{
   switch (cpkt) {
      case 130: pkt2send[0] =TRUE; break;
      case 132: pkt2send[1] =TRUE; break;
      case 134: (byte)(pkt2send[2]) =(byte)(pkt2send[2]) | 1; break;
      case 135: (byte)(pkt2send[2]) =(byte)(pkt2send[2]) | 2; break;
      case 136: (byte)(pkt2send[2]) =3; break;
      case 142: pkt2send[3] =TRUE; break;
      case 150: pkt2send[4] =TRUE; break;
      case 153: pkt2send[5] =TRUE; break;
   }
}

void pkt_sender_handler()
//
// Provvede alla trasmissione dei pacchetti prelevando i dati dalle varie
// locazioni di memoria in base al flag attivato.
//
 
{
   byte i,j;
    
   if (awt_ack) return;
   for(i=0;i<sizeof(pkt2send);i++) { 
      if (pkt2send[i]) {
         switch (i) { 
            case 0: 
               last_pkt.code = 130;
               last_pkt.len = 0;
               send_packet(130,0,NULL);  
               break;
                
            case 1: 
               last_pkt.data[0] = WR;
               last_pkt.data[1] = WC;
               last_pkt.data[2] = WL;
               last_pkt.data[3] = WA;
               last_pkt.code = 132;
               last_pkt.len = 4;
               send_packet(132,4,last_pkt.data);
               break;
            case 2: 
               switch ((byte)pkt2send[2]) {
                  case 1:
                     upd_control_buf();
                     memcpy(last_pkt.data,control_buf,8);
                     last_pkt.code = 134;
                     last_pkt.len = 8;
                     send_packet(134,8,control_buf);
                     break;
                  case 2:
                     if (video_buf_len!=40) {
                        video_buf[video_buf_len*2] = 186;
                        video_buf[video_buf_len*2+1] = 0;
                        if (video_buf_len<39) {
                           for(j=video_buf_len+1;j<40;j++) {
                              video_buf[j*2] = 32;
                              video_buf[j*2+1] =0;
                           }
                        }
                     }
                     memcpy (last_pkt.data,video_buf,80);
                     last_pkt.code = 135;
                     last_pkt.len = 80;
                     send_packet(135,80,last_pkt.data);
                     break;
                  case 3: 
                     if (video_buf_len!=40) {
                        video_buf[video_buf_len*2] = 186;
                        video_buf[video_buf_len*2+1] = 0;
                        if (video_buf_len<39) {
                           for(j=video_buf_len+1;j<40;j++) {
                              video_buf[j*2] = 32;
                              video_buf[j*2+1] =0;
                           }
                        }
                     }
                     memcpy(last_pkt.data,video_buf,80);
                     upd_control_buf();
                     memcpy(&last_pkt.data[80],control_buf,8);
                     last_pkt.code = 136;
                     last_pkt.len = 88;
                     send_packet(136,88,last_pkt.data);
                     break;
               }
               break;
            case 3:
               last_pkt.data[0] = 0xFF;
               last_pkt.data[1] = n_col;
               last_pkt.data[2] = n_lines;
               last_pkt.data[3] = n_cons;
               last_pkt.code =142;
               last_pkt.len =4;
               send_packet(142,4,last_pkt.data);
               break;
            case 4:
               last_pkt.code = 150;
               last_pkt.len = 0;
               send_packet(150,0,NULL);  
               break;
            case 5:
               last_pkt.code = 153;
               last_pkt.len = 0;
               send_packet(153,0,NULL);  
               break;
         }
         time_out_ack = 1000/LOOPTIME;
         pkt2send[i] = FALSE;
         tx_state = 1;
         return;
      }
   }
}

bool page_col_changed()
// 
// Verifica se si sono modificati la console attiva o il numero
// di colonne del modo video. In tal caso aggiorna ControlBuf e riporta true.
{  
   int i;
   bool b;
   
   b= FALSE;
   i=get_active_console();
   if (i==-1) { return FALSE;}
   if (n_cons!=i) {n_cons=i; b=TRUE; }
   if (n_col!=scr.ncols) { n_col=scr.ncols; n_col_changed=TRUE; b=TRUE;}
   return b;
}

bool bios_cur_changed()
// 
// Verifica se e cambiata la posizione del cursore.
// In tal caso aggiorna HcX, HcY e riporta true.
{ 
   bool b;

   b=FALSE;
   if (hc_x!=scr.curx) {hc_x=scr.curx; b=TRUE;}
   if (hc_y!=scr.cury) {hc_y=scr.cury; b=TRUE;}
   return b;
}

bool sc_scanner(byte atrb,byte mask,byte mask11,byte mask01,byte mask10)
/* 
 Scansiona gli attributi della pagina video e li paragona con ScBuffer.
 Tiene conto di Atrb e Mask per determinare quale attributo considerare.
 Vengono passate le maschere per determinare quali bit testare a seconda del
 soft-cursor abilitato.
 Soft-Cursor1 Bit0,Bit1
 Soft-Cursor2 Bit2,Bit3
 Soft-Cursor3 Bit4,Bit5

 Aggiorna ScBuffer con i seguenti valori:
 00 : attributo diverso da quello specificato.
 01 : attributo trovato in posizione nuova.
 10 : attributo trovato in posizione vecchia.
 Riporta true se la pagina video si  stabilizzata, cio non sono stati 
 scritti nuovi valori "01".
*/  
{
   int i;
   byte j;
   bool ret;
   int k=1;   
   
   ret=TRUE;
   for (i=0;i<sc_buflen;i++,k+=2) {
      if (((scr.data[k] ^ atrb) & mask)!=0) {
         j = mask11 ^ 0xFF;
         sc_buffer[i] &=j;
      } else {
         j=sc_buffer[i] & mask11;
         if (j==0) {
            sc_buffer[i] |= mask01;
            ret = FALSE;
         } else {
            if ((j ^ mask11)!=0) {
               if ((j & mask01)!=0) {
                  sc_buffer[i] ^= mask11;
               }
            }
         }
      }
   }
   return ret;
}

void sc_reset(byte mask11,byte mask01,byte mask10)
// 
// Scansiona ScBuffer e converte i valori "10" in valori "11".
// Il valore "11" indica che la posizione dell'attributo (stabile)  gi
// stata analizzata. 
{
   int i;
   byte j;
   for (i=0;i<sc_buflen;i++) {
      j=sc_buffer[i] & mask11;
      if (j!=0) {
         if (((j ^ mask11)!=0) && ((j & mask10)!=0)) 
            sc_buffer[i] ^= mask11;
      }
   }
}

void sc_find (word StOfs, word* ScStart, word* ScEnd, byte mask11)
// 
// Scansiona ScBuffer per determinare posizione iniziale e finale del
// soft-cursor. La scansione inizia dall'offset indicato in StOfs.
// Nel caso in cui non venga trovato il soft-cursor, ScStart riporta sempre
// 0xFFFF mentre ScEnd riporta 0xFF01 se l'attributo  presente ma fermo,
// 0xFF00 se l'attributo non  presente.
{
   int i;
   byte j;

   *ScStart=0xFFFF;
   *ScEnd=0xFF00;
   for (i=StOfs;i<sc_buflen;i++) {
      j=sc_buffer[i] & mask11;
      if (j!=0) {
         if (((j ^ mask11)!=0) && (*ScStart==0xFFFF)) {
            *ScStart=i;
         } else {
            *ScEnd=0xFF01;
         }
      } else {
         if (*ScStart!=0xFFFF) {
            *ScEnd=i-1;
            break;
         }
      }
   }
}

bool sc_bktk(byte idx, byte atrb, byte mask, byte* posidx)
// 
// Esegue il back-tracking sulle ultime posizioni del soft-cursor.
// Al termine ScPosIdx riporta la posizione del soft-cursor e la funzione
// riporta true. Se nessuna posizione valida viene trovata, riporta false.   
//
{
   int i,j;
   word pos_at;
   for (j=0;j<128;j++) {
      pos_at=sc_pos[idx][*posidx];
      pos_at*=2;
      pos_at++;
      if (((scr.data[pos_at] ^ atrb) & mask)==0) {
         pos_at--;
         *posidx=pos_at/2;
         return TRUE;
      }
      *posidx--;
      if ((!*posidx)) return FALSE;
   }
   return FALSE;
}

byte soft_cursor()
// 
// Gestione completa del Soft-Cursor. 
// Riporta i seguenti valori:
// 0 : la posizione del soft-cursor non  cambiata.
// 1 : la posizione  cambiata e OcX, OcY sono stati aggiornati.
// 2 : la posizione del soft-cursor  indefinita.
// 3 : analisi non effettuata o attributi non stabili.
//
{
   byte mask[3][3] = {0x03,0x01,0x02,0x0C,0x04,0x08,0x30,0x10,0x20};
 
   word fa,la,scposition;
   bool ok,ok1;
   byte i,j;
   byte x[3],y[3];
   bool scan[3],sc[3];
   word scendsv[3];
   byte ret;

   ret=3;
    // ok, puoi procedere 
    // bisogna attendere che le variazioni di attributo si stabilizzino 

   for(i=0;i<3;i++) {
      sc[i]=TRUE;
      scendsv[i]=0xFFFF;
      if (sc_act[i]) {
         scan[i]=sc_scanner(sc_atrb[i],sc_mask[i],mask[i][0],mask[i][1],mask[i][2]);
         sc[i]=scan[i];
      } else scan[i]=FALSE;
   }
   if (!(sc[0] && sc[1] && sc[2])) {
       sc_scanact =TRUE;
       return ret;
   } else sc_scanact=FALSE;

   ok1=FALSE;
   for(i=0;i<3;i++) {
      if (scan[i]) {
         ok1=TRUE;
         fa=y_min*n_col+x_min;
         la=y_max*n_col+x_max;
         if ((((word)sc_row1[i]-1)*(word)n_col)>fa) fa=(word)(sc_row1[i]-1)*(word)n_col;
         if (((word)sc_row2[i]*(word)n_col-1)<la) la=(word)sc_row2[i]*(word)n_col-1;
         ok=FALSE;
         do {
            sc_find(fa,&sc_start,&sc_end,mask[i][0]);
            scendsv[i]=sc_end;
            if (sc_start==0xFFFF) {
               ok=TRUE; 
            } else {
               if (sc_first[i]) scposition=sc_start; else scposition=sc_end;
               if (scposition>la) {
                  ok=TRUE;
                  sc_start=0xFFFF;
                  sc_end=0xFF01;
               } else {
                  sc_x[i]=(scposition % n_col);
                  sc_y[i]=(scposition/n_col);
                  if ((sc_x[i]>=x_min) && (sc_x[i]<=x_max) && 
                      (sc_y[i]>=y_min) && (sc_y[i]<=y_max))      
                     ok=TRUE;
                  else 
                     fa=sc_end+1;
               }
            }
         } while (!ok);         
         sc_reset(mask[i][0],mask[i][1],mask[i][2]);
         if (sc_start!=0xFFFF) break;
      }
   }
   if (!ok1) return ret;
   
   if (sc_start!=0xFFFF) {
      // trovata nuova posizione, aggiorna ScPos, ScX, ScY 
      sc_posidx[i]=(sc_posidx[i]+1) & 0x7F;
      if (sc_first[i])  
         sc_pos[i][sc_posidx[i]]=sc_start;
      else
         sc_pos[i][sc_posidx[i]]=sc_end;
      sc_x[i]=(sc_pos[i][sc_posidx[i]] % n_col);
      sc_y[i]=(sc_pos[i][sc_posidx[i]] / n_col);
      if ((oc_x!=sc_x[i]) || (oc_y!=sc_y[i])) {
         oc_x=sc_x[i];
         oc_y=sc_y[i];
         scnum =i;
         ret=1;
       } else ret=0;
       return ret;
   }
   
   j=0;
   for(i=0;i<3;i++) {
      if ((sc_act[i]) && (scendsv[i]==0xFF01)) {
         // l'attributo  comunque ancora presente, tenta il back-tracking 
         if (sc_bktk(i,sc_atrb[i],sc_mask[i],&sc_posidx[i])) {
            // trovata vecchia posizione } 
            x[i]=(sc_pos[i][sc_posidx[i]] % n_col);
            y[i]=(sc_pos[i][sc_posidx[i]] / n_col);
            if ((sc_x[i]==x[i]) && (sc_y[i]==y[i]))
               j=j | (1 << i);
            else 
               j=j | (8 << i);
         }
      }
   }
   if (j==0) return 2;

   ret=0;
   ok=FALSE;

   repeat_btrk:  
   for(i=0;i<3;i++) {
      if ((j & (1 << i))==0) {
         if ((j & (8 << i))!=0) {
            if (ok || ((oc_x==sc_x[i]) && (oc_y==sc_y[i]))) {
               sc_x[i]=x[i]; oc_x=x[i];
               sc_y[i]=y[i]; oc_y=y[i];
               ret=1;
               scnum =i;
               return 1;
            } else {
               sc_x[i]=x[i];
               sc_y[i]=y[i];
               j=j ^ (9 << i);
            }
         } else {
            if ((i==scnum) && (j<8)) {
               ok=TRUE;
               j=j << 3; 
               goto repeat_btrk;
            }
         }
      }
   }
   return ret;
}

void update_sc_par()
// 
// Aggiorna i parametri del soft-cursor.
{
   int i,j;
   sc_tracking=pkt.data[0];
   for(i=0;i<3;i++) {
      sc_atrb[i]=pkt.data[i+1];
      sc_mask[i]=pkt.data[i+4];
      sc_first[i]=pkt.data[i+7];
      hc_tracking[i]=pkt.data[i+10];
      sc_row1[i]=pkt.data[i+13];
      sc_row2[i]=pkt.data[i+16];
      sc_act[i]=pkt.data[i+19];
      for (j=0;j<128;j++) sc_pos[i][j]=0;
   }
   for (i=0;i<4000;i++) sc_buffer[i]=0;
}

bool control_changed1()
// 
// Compara i vari parametri di controllo con ControlBuf.
// Se diversi aggiorna ControlBuf e ritorna con true.
// La lettura della posizione del cursore puo avvenire in 2 modi diversi:
// - Attraverso i valori riportati dal device vcsa0.
// - Utilizzando l'algoritmo del Soft-Cursor.
//
{
 bool chg1,chg2,ok;
 byte j,i;
   chg1=page_col_changed();
   if (!sc_tracking) {
       // Tracking Hard-Cursor
       chg2=bios_cur_changed();
       if (chg2) { cur_x=hc_x; cur_y=hc_y; }
   } else {
       // Tracking Soft-Cursor
      chg2=FALSE;
      i=soft_cursor();
      
      if (i<3) {
         for(j=0;j<3;j++) {
            if ((sc_act[j]) && (hc_tracking[j]))  
            //  controlla l'Hard-Cursor 
               chg2=bios_cur_changed();
            if (chg2) break;
         }
      }
      switch (i) {
         case 3: // analisi non effettuata o attributo instabile }
            chg2=FALSE;
            break;
         case 0: // non si  spostato 
            ok=FALSE;
            for (j=0;j<3;j++) {
               if (sc_act[j] && hc_tracking[j] && chg2) {
                  cur_x=hc_x; 
                  cur_y=hc_y;
                  ok=TRUE;
                  break;
               }
            }
            chg2=ok;
            break;
         case 1: // si  spostato, aggiorna CurX, CurY 
            chg2=TRUE;
            cur_x=oc_x;
            cur_y=oc_y;
            break;   
         case 2: // si  perduto, riprendi l'Hard-Cursor 
            bios_cur_changed();
            chg2=((cur_x!=hc_x) || (cur_y!=hc_y));   
            cur_x=hc_x;
            cur_y=hc_y;
            oc_x=hc_x;
            oc_y=hc_y;
            break;
       }            
   }
   
   if (get_nlines() | (chg1 | chg2)) {
      upd_control_buf();
      return TRUE;
   }
   return FALSE;
}

bool control_changed()
// 
// Analoga alla precedente ma attende la stabilizzazione della variazione. 
//
{
   if (control_changed1()) { cc=trk_delay; cc_flag=TRUE;}
   if (cc_flag && (cc==0)) { cc_flag=FALSE; return TRUE;} else return FALSE;
}

void set_mb_position(char delta, char dir, bool* ok)
/* 
 Setta la posizione di MB408 in base allo spostamento relativo indicato.
 Lo spostamento Delta e espresso in linee o caratteri in funzione del
 valore di Dir. Ok riporta false se lo spostamento richiesto non e
 effettuabile. L'area attiva di spostamento e definita dalle coordinate
 Xmin,Ymin e Xmax,Ymax (0,0 e 79,24 per un full screen a 80x25).
 Calcola anche il valore della variabile globale StartAdrs che riporta
 l'offset iniziale del buffer di memoria video su cui  posizionato MB408.
*/
{
   *ok=FALSE;
   // Se MB408  fuori dalla window definita allora riportalo nell'angolo in
   // alto a sinistra della window.
   // Nota: la window deve essere larga almeno 40 caratteri.}
   if ((mb_x<x_min) || (mb_x>x_max) || (mb_y<y_min) || (mb_y>y_max)) {  
      mb_x=x_min;                                                       
      mb_y=y_min;                                                       
      *ok=TRUE;                                                         
      start_adrs=(mb_y*n_col+mb_x)*2;                                   
      return;                                                           
   }                                                                    
   // Se Delta=0 ricalcola solo StartAdrs       
   if (delta==0) {                              
      *ok=TRUE;                                 
      start_adrs=(mb_y*n_col+mb_x)*2;           
      if ((mb_x>x_min+39) || ((x_max-mb_x)<40))  
         video_buf_len=x_max-mb_x+1;              
      else                                      
         video_buf_len=40;                      
      return;                                   
   }                                            
   switch (dir) {
      case up_v: 
         if (mb_y==y_min) return;
         mb_y-=delta;
         if (mb_y<y_min) mb_y=y_min;
         break;
      case down_v: 
         if (mb_y==y_max) return;
         mb_y+=delta;
         if (mb_y>y_max) mb_y=y_max;
         break;
      case left_h: 
         if (mb_x==x_min) {
            if (mb_y==y_min) return;
            mb_y--;
            mb_x=x_min+mb_ofs+x_step;
            if ((x_max-(mb_x+mb_ofs))>x_step)  mb_x+=x_step;
            if (mb_x>=x_max) mb_x=x_min+mb_ofs;
            video_buf_len=x_max-mb_x+1;
         } else {
            mb_x-=delta;
            if (mb_x<x_min) mb_x=x_min;
            if ((mb_x>(x_min+39)) || ((x_max-mb_x)<40)) 
               video_buf_len=x_max-mb_x+1; 
            else 
               video_buf_len=40;
         }
         break;
      case right_h: 
         if (x_max<=(mb_x+40)) {
            if (mb_y==y_max) return;
            mb_y++;
            mb_x=x_min;
            video_buf_len=40;
         }  else {
            mb_x+=delta;
            if (x_max>(mb_x+40)) 
               video_buf_len=40;
            else
               video_buf_len=x_max-mb_x+1;
         }
         break;
   }
   *ok=TRUE;
   start_adrs=(mb_y*n_col+mb_x)*2;
}

bool video_changed1 (int st_adrs)
// 
// Compara l'area video che inizia con VideoBuf.
// Se diversi aggiorna VideoBuf e ritorna con true.
//
{
   
   int i;
   bool b;

   b=FALSE;
   for (i=0;i<video_buf_len*2;i++) {
      if (scr.data[i+st_adrs]!=video_buf[i]) {
         b=TRUE;
         break;
      }
   }
   if (b) {
      for (i=0;i<video_buf_len*2;i++) video_buf[i]=scr.data[i+st_adrs];
   }
   return b;
}

bool video_changed(int st_adrs)
// 
// Analoga alla precedente ma attende la stabilizzazione della variazione.
//
{
   if (cc_flag || sc_scanact) return FALSE;
   return (video_changed1(st_adrs));
}

void track_cursor(byte sidechar)
/* 
 Esegue la procedura di aggancio con il cursore video. Se il cursore  fuori
 dalla finestra attualmente definita, allora non si procede.
 Sidechar indica il numero minimo di caratteri a destra o sinistra del cursore
 che deve sempre esserci.
*/

{
   byte leftedge,rightedge;

   if ((cur_y<y_min) || (cur_y>y_max) || (cur_x<x_min) || (cur_x>x_max)) return;
   mb_y=cur_y;
   mb_x=x_min+40;
   ok=FALSE;
   do {
      if (mb_x==x_min) leftedge=mb_x; else leftedge=mb_x+sidechar;
      if (mb_x>(x_min+39)) rightedge=x_max; else rightedge=mb_x+39-sidechar;
      if ((cur_x>=leftedge) && (cur_x<=rightedge)) ok=TRUE; 
        else {
          if ((mb_x>(x_min+39)) && (right_step!=0)) 
             mb_x-=right_step;
          else 
             mb_x-=x_step;
          if (mb_x<x_min) mb_x=x_min;
        }
   } while (!ok);
   set_mb_position(0,up_v,&ok);
}

void track_the_cursor()
// 
// Utilizza la procedura precedente tenendo anche conto del modo di aggancio
// con il cursore (TrkType40).
//
{
   if (trk_type40) {
      x_step=40; 
      track_cursor(0);
   } else {
      x_step=36;   
      track_cursor(2);
   }
}

void set_new_videomode()
// 
// Setup necessario quando si cambia modo video.
//
{
   get_nlines();
   if (full_screen) {
      x_min = 0;
      x_max = scr.ncols-1;
      y_min = 0;
      y_max = n_lines-1;
   } else {
      x_min=x_min0;
      y_min=y_min0;
      x_max=x_max0;
      x_max=y_max0;
      if (!((x_min!=0) && (x_min<scr.ncols))) x_min = 0;
      if (!(x_max<scr.ncols)) x_max = scr.ncols-1;
      if (!((y_min!=0) && (y_min<n_lines))) y_min = 0;
      if (!(y_max<n_lines)) y_max = n_lines-1;
   }
   mb_x = -1;
   mb_y = -1;
   mb_ofs = 0;
   ad_x = 0;
   ad_y = 0;
   get_side_steps();
   set_mb_position(0,up_v,&ok);
   video_changed1(start_adrs);
   control_changed1();  
   if (tracking) track_the_cursor();
   n_col_changed = FALSE;
   n_line_changed = FALSE;
   sc_buflen = scr.ncols*n_lines;
}

void upd_setup_par()
// 
// Aggiorna i parametri di setup.
//
{
   int i,j;
   tracking = pkt.data[0];
   trk_type40 = pkt.data[1];
   trk_delay = pkt.data[2]*2;
   sc_tracking = pkt.data[3];
   for(i=0;i<3;i++) {
      sc_atrb[i]=pkt.data[i+4];
      sc_mask[i]=pkt.data[i+7];
      sc_first[i]=pkt.data[i+10];
      hc_tracking[i]=pkt.data[i+13];
      sc_row1[i]=pkt.data[i+16];
      sc_row2[i]=pkt.data[i+19];
      sc_act[i]=pkt.data[i+22];
      for(j=0;j<127;j++) sc_pos[i][j]=0;
   }
   if (trk_type40) x_step=40; else x_step=36;
   for (i=0;i<4000;i++) sc_buffer[i]=0;
}

void move_mb_hor (char dir, bool* ok)
// 
// Sposta MB408 a destra o a sinistra dello step indicato da Xstep.
// Ok indica se lo spostamento  stato possibile.

{
   switch (dir) {
      case left_h: 
         if (mb_x==x_min) {
            set_mb_position (1,dir,ok); 
         } else {
            if ((ofs_step) && (mb_x<=x_min+x_step)) {
               set_mb_position(ofs_step,dir,ok);
            } else {
               set_mb_position(x_step,dir,ok);
            }
         }
         break;
      case right_h: 
      if ((ofs_step) && (mb_x==x_min)) {
         set_mb_position(ofs_step,dir,ok); 
      } else {
         if (mb_x>=(x_max-39)) {
            set_mb_position(1,dir,ok); 
         } else {
            set_mb_position(x_step,dir,ok);
         }
      }
      break;
   }
}


/***********************************************************
PROCEDURE PER LA GESTIONE DEGLI SPOSTAMENTI IN AUTOREAD ***
***********************************************************/


bool rg_handler (bool* ok)
/*
 Gestione dello spostamento a destra in modo Autoread.
 Il parametro "ok" indica se  stato effettuato lo spostamento della
 finestra di lettura.
 Riporta True se deve saltare la gestione dello spostamento.
*/

{
   if (type_auto_read==1) {
      if (!(*ok)) {
         stuf_key_str(1,DNAR);
         mb_x=x_min;
         set_mb_position(0,up_v,ok);
         n_cons=0xFF;
         send_ack();
         return TRUE;
      }
   } else {
      if (mb_x==x_min) {
         stuf_key_str(1,DNAR);
      } else {
         if (!(*ok)) {
            stuf_key_str(1,DNAR);
            mb_x=x_min;
            set_mb_position(0,up_v,ok);
            n_cons=0xFF;
            send_ack();
            return TRUE;
         }
      }
   }
   return FALSE;
}

bool lf_handler (bool *ok)
/*
 Gestione dello spostamento a sinistra in modo Autoread.
 Il parametro "ok" indica se  stato effettuato lo spostamento della
 finestra di lettura.
 Riporta True se deve saltare la gestione dello spostamento.
*/
{
   if (type_auto_read==1) {
      if (!(*ok)) {
         stuf_key_str(1,UPAR);
         mb_x=x_max-39;
         set_mb_position(0,up_v,ok);
         n_cons=0xFF;
         send_ack();
         return TRUE;
      }
   } else {
      if (mb_x==(x_max-39)) {
         stuf_key_str(1,UPAR);
      } else {
         if (!(*ok)) {
            stuf_key_str(1,UPAR);
            mb_x=x_max-39;
            set_mb_position(0,up_v,ok);
            n_cons=0xFF;
            send_ack();
         }
      }
   }
   return FALSE;
}

void up_handler (bool* ok)
/*
 Gestione dello spostamento verso l'alto in modo Autoread.
 Il parametro "ok" indica se e stato effettuato lo spostamento della
 finestra di lettura.
*/
{
   if (type_auto_read==1) {
      if (!(*ok)) {
         stuf_key_str(1,UPAR);
      }
   } else {
      stuf_key_str(1,UPAR);
   }
}

void dn_handler (bool *ok)
/*
 Gestione dello spostamento verso il basso in modo Autoread.
 Il parametro "ok" indica se  stato effettuato lo spostamento della
 finestra di lettura.
*/
{
   if (type_auto_read==1) {
      if (!(*ok)) {
         stuf_key_str(1,DNAR);
      }
   } else {
      stuf_key_str(1,DNAR);
   }
}

void search_text (char dir, byte filter,bool* found)
// 
// Si sposta in avanti o indietro alla ricerca del prossimo testo.
//
{
   byte j,c;
   bool ok;
    
   ok=TRUE;
   while (ok) {
      video_changed1(start_adrs);
      *found=FALSE;
      for (j=0;j<video_buf_len;j++) {
          c=video_buf[j*2];
          if ((c>0) && (c<=filter) && (c!=32)) *found=TRUE;
      }
      if (*found) ok=FALSE; else move_mb_hor(dir,&ok);
    }
}

void align()
// 
// Allinea MB408 in base all'offset MbOfs attuale.
//
{
   byte newmbx,d,i;
   newmbx=x_min;
   i=x_min;
   d=mb_x-x_min;
   do {
      if ((newmbx==x_min) && (ofs_step!=0)) 
         newmbx+=ofs_step;
      else 
         newmbx+=x_step;
      if (newmbx>(x_max-39)) newmbx=x_min+40;
      if (((byte)(mb_x-newmbx))>=d) {
         mb_x=i;
         return;
      } else {
         i=newmbx;
         d=(byte)(mb_x-newmbx);
      }
   } while (!(newmbx>=(x_min+40)));
   mb_x=i;
}

void set_hortab (char dir, char* tpos)
// 
// Gestione tabulazione orizzontale. Dir indica la direzione nella ricerca della
// tabulazione (destra o sinistra) e tpos riporta il valore della tabulazione
// trovata (tpos=0xFF se non viene trovata alcuna tabulazione).
{
   byte tabpos[10];
   byte i,j,t;
   for(i=0;i<10;i++) tabpos[i]=pkt.data[i+1];
   j=0;
   if (dir==right_h) {
      *tpos=100;
      for(i=0;i<10;i++) {
         t=tabpos[i];
         if ((t<128) && (t>mb_x) && (t<=x_max))
            if (t<(*tpos)) {
               *tpos=t;
               j=i+1;
            }
      }
   } else {
      *tpos=-1;
      for(i=0;i<10;i++) {
         t=tabpos[i];
         if ((t<128) && (t<mb_x) && (t>=x_min))        
            if (t>(*tpos)) {
               *tpos=t;
               j=i+1;
            }
      }      
   }
   if (j==0) *tpos=-1;
}

void find_window (byte* R,byte* C,byte* L,byte* A)
// 
// Partendo dalla posizione indicata da ADx e ADy ricerca una finestra sullo
// schermo, riportando R=riga iniziale, C=colonna iniziale, L=larghezza,
// A=altezza. Se non viene trovata alcuna finestra riporta 4 valori 0.
//
{
   byte windchar[4] = {218,191,217,192}; 
   byte vb;
   char x,y,res;
    
   void moveforward(byte* vb1)
   {
      ad_x++;
      if (ad_x>(n_col-1)) { 
         ad_x=0; ad_y++;
         if (ad_y>(n_lines-1)) ad_y=0;
      }
      *vb1=scr.data[(word)((word)ad_y*(word)n_col+(word)ad_x)*2];
   }   

   void scan(char x,char y,char dir, byte wchar, char* res)
   // 
   // Scansiona lo schermo partendo dalla posizione x,y nella direzione dir,
   // ricercando uno dei caratteri finestra del gruppo indicato da wchar. Se viene
   // trovato uno di tali caratteri, la sua posizione  in res altrimenti res=-1.
   // wchar = 0  ---> angolo in alto a sinistra
   // wchar = 1  ---> angolo in alto a destra
   // wchar = 2  ---> angolo in basso a destra
   // wchar = 3 ---> angolo in basso a sinistra.
   {
      byte b;
      switch (dir) {
         case up_v: 
            while (y>=0) {
               b=scr.data[(word)((word)y*(word)n_col+(word)x)*2];
               if ((b==windchar[wchar])) {
                  *res=y;
                  return;
               }
               y--;
            };
            *res=-1;
            return;
                      
         case right_h: 
            while (x<n_col) {
               b=scr.data[(word)((word)y*(word)n_col+(word)x)*2];
               if ((b==windchar[wchar])) {
                  *res=x;
                  return;
               }
               x++;
            }
            *res=-1;
            return;
      }
   }
                    

   old_x=ad_x;
   old_y=ad_y;
   *R=0; *C=0; *L=0; *A=0;
   vb=scr.data[(word)((word)ad_y*(word)n_col+(word)ad_x)*2];
   do {
      if (vb==192) {
         // Hai trovato un vertice in basso a sinistra di una finestra 
         *R=ad_y;
         *C=ad_x;  
         // Scan a destra per vedere quanto  larga.
         scan(ad_x,ad_y,right_h,2,&res);
         if (res>=0) *L=res-ad_x+1;
         // Scan in alto per vedere quanto  alta.
         scan(ad_x,ad_y,up_v,0,&res);
         if (res>=0) *A=ad_y-res+1;
         if ((L==0) && (A!=0)) {
             // Riprova a trovare la larghezza partendo da in alto a sinistra 
            scan(ad_x,ad_y-(*A)+1,right_h,1,&res);
            if (res>=0) *L=res-ad_x+1;
         }
         if ((L!=0) && (A==0)) {
            // Riprova a trovare l'altezza partendo da in basso a destra }
            scan(ad_x+(*L)-1,ad_y,up_v,1,&res);
            if (res>=0) *A=ad_y-res+1;
         }
         if ((L!=0) && (A!=0)) {
            // La finestra  definita completamente 
            *R=*R-(*A)+1;
            moveforward(&vb);
            return;
         }
         // Ricerca della finestra fallita, continua 
      }
      moveforward(&vb);
    } while (!(((old_x==ad_x) && (old_y==ad_y)))); 
}

void rx_crpar()
// 
// Riceve i parametri di setup per il cursor routing.
//
{
   cur_routex=pkt.data[1];
   cur_routey=pkt.data[2];
   if (tracking) route_cur=2; else route_cur=1;
   tracking=FALSE;
   y_ok=TRUE;
   arr_routey=cur_y-cur_routey;
   arr_routex=0;
   route_retry=MAXROUTERETRY;
   route_state=2;
}

void cursor_routing()
//
// gestione del cursor routing
//
{
   byte i;

   if (route_cur==0) return;
   
   if (arr_routey!=0) {
      switch (route_state) {
         case 2: 
            i=abs(arr_routey);
            if (i>1) {
               i=2;
               if (arr_routey<0) 
                  next_pos=cur_y+2; 
               else 
                  next_pos=cur_y-2;
            }
            if (arr_routey<0) { 
               arr_routey += i; 
               stuf_key_str(i,DNAR); 
            } else { 
               arr_routey -= i; 
               stuf_key_str(i,UPAR); 
            }
            route_state--;
            break;
         case 1: 
            i=scr.cury;
            if (i==next_pos) {
               fast_cur=TRUE; 
            } else {
               fast_cur=FALSE;
               arr_routey=arr_routey+((int)(i)-next_pos);
            }
            route_state--;
            break;  
         case 0: 
            if (fast_cur) {
               i=abs(arr_routey);
//!!!               if (i>15) i=15;
            } else i=1;
            if (arr_routey<0) { 
               arr_routey+=i; 
               stuf_key_str(i,DNAR);
            } else {
               arr_routey-=i; 
               stuf_key_str(i,UPAR); 
            }
            break;
      }
   } else {
      if (cur_y==cur_routey) {
         if (y_ok) { 
            arr_routex=cur_x-cur_routex; 
            y_ok=FALSE;
            route_state=2;
         }
         if (arr_routex!=0) {
            switch (route_state) {
               case 2:
                  i=abs(arr_routex);
                  if (i>1) {
                     i=2;
                     if (arr_routex<0) 
                        next_pos=cur_x+2;
                     else 
                        next_pos=cur_x-2;
                  }   
                  if (arr_routex<0) { 
                     arr_routex += i; 
                     stuf_key_str(i,RIGHTAR); 
                  } else { 
                     arr_routex -= i; 
                     stuf_key_str(i,LEFTAR); 
                  }
                  route_state--;
                  break;
               case 1: 
                  i=scr.curx;
                  if (i==next_pos) {
                     fast_cur=TRUE;
                  } else {
                     fast_cur=FALSE;
                     arr_routex=arr_routex+((int)(i)-next_pos);
                  }
                  route_state--;
                  break;  
               case 0:
                  if (fast_cur) {
                     i=abs(arr_routex);
//!!!                     if (i>15) i=15;
                  } else i=1;
                  if (arr_routex<0) { 
                     arr_routex += i; 
                     stuf_key_str(i,RIGHTAR); 
                  } else {
                     arr_routex -= i; 
                     stuf_key_str(i,LEFTAR); 
                  }
                  break;
            }
         } else {
            if (route_cur==2) tracking=TRUE;
            route_cur=0;
         }
      } else {
         if (route_retry==0) {
            if (route_cur==2) tracking=TRUE;
            route_cur=0; 
         } else route_retry--;
      }
   }
}

//
// MAIN PROGRAM
//

main(int argc, char *argv[])
{
   // Scriviamo il titolo

   puts(TITLE);

   // Controlliamo se il programma  gi in esecuzione

   switch (program_loaded()) {
      case -1: puts("Errore nell'esecuzione del comando \"pstree -p \"");
               exit(1);
      case  0: break;
      default: printf("Programma gi in esecuzione - PID Number %s \n",pidstr);
               exit(1);
   }

   // Leggiamo la linea comandi

   parse_options(argc,argv);

   // Apriamo la console virtuale
   vcfd=open(VCSADEV,O_RDONLY);
   if (vcfd==-1) {
      printf("Errore di apertura della console virtuale (%s)\n",VCSADEV);
      close_program();
      exit(1);
   }

/*   
   // Apriamo il file log
   log=open("./logfile",O_RDWR | O_CREAT | O_TRUNC);

   if (log==-1) {
      printf("Errore di apertura logfile\n");
      close_program();
      exit(1);
   } else printf("Apertura logfile\n");
   lseek(log,0,SEEK_SET);
*/

   // Apriamo la porta seriale

   port_ok=(open_serial()==0);
   if (!port_ok) {
      printf("Errore di apertura della porta seriale %s (%s)\n",dev_str[what_com],com_str[what_com]);
      close_program();
      exit(1);
   }

   // Eseguiamo il fork del processo

   switch (fork()) {
      case -1: // errore nel fork
         puts("Errore nell'installazione del programma");
         close_program();
         exit(1);
      case 0:  // processo child, che diventa un daemon
         close(STDIN_FILENO);
         close(STDOUT_FILENO);
         close(STDERR_FILENO);
         if (setsid()==-1) {
            close_program();
            exit (1);
         }
         break;
      default: // processo parent, termina subito
         printf("MB408 collegato a %s (%s)\n",dev_str[what_com],com_str[what_com]);
         return 0;
   }

   // Processo child: installiamo gli handler per i segnali

   if (signal(SIGTERM,SIGTERM_handler)==SIG_IGN) signal(SIGTERM,SIG_IGN);
   signal(SIGINT,SIG_IGN);
//   signal(SIGKILL,SIGKILL_handler);
   signal(SIGHUP,SIG_IGN);

   // Disabilita il paging della memoria e impostiamo il task con la 
   // massima priorit del gruppo 0

   mlockall(MCL_CURRENT);
//   sched_getparam(0,&sp);
//   sp.sched_priority=1;
//   sched_setscheduler(0,SCHED_FIFO,&sp);
   setpriority(PRIO_PROCESS,0,PRIO_MIN);
   
   init();
   
   while (keep_going) {
      read_screen();
      rx_pkt_handler();
      if (prog_status==START) {
         if (pkt_avl) {
            // in attesa di un pacchetto parametri di setup iniziale 
            if (pkt.code==129) {
               prog_status = RUN;
               beep(1000,100);
            } else {
               pkt_avl = FALSE;
               send_ack();
            }
         } else {
            // ogni secondo richiedi i parametri di setup iniziale }
            if (t_sync) t_sync--; else {
               t_sync = 1000/LOOPTIME;
               set_pkt(130);
               awt_ack = FALSE;
            }
         }
      } else {
         // contatore per la stabilizzazione del cursore 
         if (cc>0) cc--;
         
         // controllo se sto aspettando un pacchetto di conferma ricezione OK 
         if (awt_ack) {
            time_out_ack--;
            if (time_out_ack==0) {
               if (tx_state>4) {
                  reset_prog_state();
                  goto exit_timer;
               }
               time_out_ack = 1000/LOOPTIME;
               tx_state++;
               send_packet(last_pkt.code,last_pkt.len,last_pkt.data);
            }
            goto exit_timer;
         }
         if (pkt_avl) {
            pkt_avl=FALSE;
            switch (pkt.code) {
               case 129:  // parametri di setup iniziale
                  upd_setup_par();
                  set_new_videomode();
                  auto_read=FALSE;
                  set_pkt(136);
                  prog_status=RUN;
                  break;
               case 143: // richiesta di spostamento 
                  switch (pkt.data[0]) {
                     case RG: 
                        // a destra di 40 caratteri 
                        move_mb_hor(right_h,&mov_ok);
                        if (auto_read) 
                           if (rg_handler(&mov_ok)) goto exit_timer;
                        if (mov_ok) video_changed1(start_adrs);
                        break;
                     case LF: // a sinistra di 40 caratteri 
                        move_mb_hor(left_h,&mov_ok);
                        if (auto_read) 
                           if (lf_handler(&mov_ok)) goto exit_timer;
                        if (mov_ok) video_changed1(start_adrs);
                        break;
                     case UP: // in alto di 1 linea 
                        set_mb_position(1,up_v,&mov_ok);
                        if (auto_read) {
                           up_handler(&mov_ok);
                           if (!mov_ok) {
                              send_ack();
                              goto exit_timer;
                           }
                        }
                        if (mov_ok) video_changed1(start_adrs);
                        break;
                     case DN: // in basso di 1 linea 
                        set_mb_position(1,down_v,&mov_ok);
                        if (auto_read) {
                           dn_handler(&mov_ok);
                           if (!mov_ok) {
                              send_ack();
                              goto exit_timer;
                           }
                        }
                        if (mov_ok) video_changed1(start_adrs);
                        break;
                     case N_RG: // avanti al prossimo testo 
                        if ((pkt.data[1]!=filter_on) && (pkt.data[1]!=filter_off))
                           mov_ok=FALSE;
                        else {
                           old_x = mb_x;
                           old_y = mb_y;
                           move_mb_hor(right_h,&mov_ok);
                           if (mov_ok) {
                              search_text(right_h,pkt.data[1],&found);
                              if (!found) {
                                 mb_x=old_x;
                                 mb_y=old_y;
                                 set_mb_position(0,up_v,&ok);
                                 mov_ok=FALSE;
                              }
                           }
                        }
                        break;
                     case N_LF: // indietro al testo precedente 
                        if ((pkt.data[1]!=filter_on) && (pkt.data[1]!=filter_off))
                           mov_ok=FALSE;
                        else {
                           old_x = mb_x;
                           old_y = mb_y;
                           move_mb_hor(left_h,&mov_ok);
                           if (mov_ok) {
                              search_text(left_h,pkt.data[1],&found);
                              if (!found) {
                                 mb_x=old_x;
                                 mb_y=old_y;
                                 set_mb_position(0,up_v,&ok);
                                 mov_ok=FALSE;
                              }
                           }
                        }
                        break;
                  }
                  if (mov_ok) set_pkt(136); else set_pkt(150);
                  break;   
               case 138: // toggle aggancio cursore 
                  tracking=(pkt.data[0]!=0);
                  n_cons=0xFF;
                  break;
               case 139: // toggle passo di avanzamento (36/40 caratteri) 
                  trk_type40=(pkt.data[0]!=0);
                  if (trk_type40) x_step=40; else x_step=36;
                  get_side_steps();
                  align();
                  if (tracking) track_the_cursor();
                  set_mb_position(0,up_v,&ok);
                  video_changed1(start_adrs);
                  set_pkt(136);
                  break;
               case 140: // toggle ritardo aggancio cursore 
                  trk_delay=pkt.data[0]*2;
                  break;
               case 141: // invia informazioni video 
                  set_pkt(142);
                  break;
               case 144: // tabulazione verticale 
                  if (pkt.data[0]==S_UP) {
                     if ((pkt.data[1]<2) || (pkt.data[1]>(n_lines-1)) || (mb_y==0)) 
                        mov_ok=FALSE;
                     else {
                        set_mb_position(pkt.data[1],up_v,&mov_ok);
                        if (mov_ok) video_changed1(start_adrs);
                     }
                  } else {
                     if ((pkt.data[1]<2) || (pkt.data[1]>(n_lines-1)) || (mb_y==(n_lines-1))) 
                        mov_ok=FALSE;
                     else {
                        set_mb_position(pkt.data[1],down_v,&mov_ok);
                        if (mov_ok) video_changed1(start_adrs);
                     }
                  }
                  if (mov_ok) set_pkt(136); else set_pkt(150);
                  break;
               case 145: // salta a inizio/fine testo 
                  if (pkt.data[0]==S_F1) {
                     if ((pkt.data[1]!=filter_on) && (pkt.data[1]!=filter_off)) 
                        mov_ok=FALSE;
                     else {
                        old_x=mb_x;
                        old_y=mb_y;
                        mb_x=x_min;
                        mb_y=y_min;
                        set_mb_position(0,up_v,&ok);
                        search_text(right_h,pkt.data[1],&found);
                        if (!found) {
                           mb_x=old_x;
                           mb_y=old_y;
                           set_mb_position(0,up_v,&ok);
                           mov_ok=FALSE;
                        } else mov_ok=TRUE;
                     }
                  } else {
                     if ((pkt.data[1]!=filter_on) && (pkt.data[1]!=filter_off))
                        mov_ok=FALSE;
                     else {
                        old_x=mb_x;
                        old_y=mb_y;
                        mb_x=x_max-39;
                        mb_y=y_max;
                        set_mb_position(0,up_v,&ok);
                        search_text(left_h,pkt.data[1],&found);
                        if (!found) {
                           mb_x=old_x;
                           mb_y=old_y;
                           set_mb_position(0,up_v,&ok);
                           mov_ok=FALSE;
                        } else mov_ok=TRUE;
                     }
                  }
                 if (mov_ok) set_pkt(136); else set_pkt(150);
                 break;
               case 149: // gestione tasti generici 
                  switch (pkt.data[0]) {
                     case S_LF: // Inserisci un Page Up nel buffer di tastiera 
                       if (route_cur==0) stuf_key(PGUP);
                       break;    

                     case S_RG: // Inserisci un Page Down nel buffer di tastiera 
                        if (route_cur==0) stuf_key(PGDN);
                        break;

                     case F1: // inizio prima riga 
                        mb_x=x_min;
                        mb_y=y_min;
                        set_mb_position(0,up_v,&ok);
                        video_changed1(start_adrs);
                        set_pkt(136);
                        break;

                     case F2: // inizio ultima riga 
                        mb_x=x_min;
                        mb_y=y_max;
                        set_mb_position(0,up_v,&ok);
                        video_changed1(start_adrs);
                        set_pkt(136);
                        break;

                      case F8: // shift a destra un carattere
                        if (mb_x<(x_max-39)) {
                           mb_ofs=mb_x-x_min+1;
                           get_side_steps();
                           set_mb_position(1,right_h,&ok);
                           video_changed1(start_adrs);
                           set_pkt(136);
                         } else set_pkt(150);
                         break;

                      case F3: // shift a sinistra un carattere 
                        if (mb_x>x_min) {
                           mb_ofs=mb_x-x_min-1;
                           get_side_steps();
                           set_mb_position(1,left_h,&ok);
                           video_changed1(start_adrs);
                           set_pkt(136);
                        } else set_pkt(150);
                        break;

                     case SL_F3: // azzera shift caratteri 
                        mb_ofs=0;
                        get_side_steps();
                        align();
                        if (tracking) track_the_cursor();
                        set_mb_position(0,up_v,&ok);
                        video_changed1(start_adrs);
                        set_pkt(136);
                        break;

                     case F4: // riaggancio cursore 
                        re_track=TRUE;
                        n_cons=0xFF;
                        break;
                     default:
                        set_pkt(150);
                        break;
                  }
                  break;
               case 147: // richiama marker 
                  if ((pkt.data[0]<x_min) || (pkt.data[0]>x_max) ||
                      (pkt.data[1]<y_min) || (pkt.data[1]>y_max)) 
                     set_pkt(150);
                  else {
                     mb_x=pkt.data[0]; 
                     mb_y=pkt.data[1];   
                     set_mb_position(0,up_v,&ok);
                     video_changed1(start_adrs);
                     set_pkt(136);
                  }
                  break;
               case 146: // tabulazione orizzontale avanti e indietro 
                  if (pkt.data[0]==S_F8) 
                     set_hortab(right_h,&k);
                  else 
                     set_hortab(left_h,&k);
                  if (k!=-1) {
                     mb_x=k;   
                     set_mb_position(0,up_v,&ok);
                     video_changed1(start_adrs);
                     set_pkt(136);
                  } else set_pkt(150);
                  break;
               case 148: // gestione Soft-Cursor
                  update_sc_par();
                  n_cons =0xFF;
                  if (sc_tracking) oc_x=0xFF; else hc_x=0xFF;
                  break;

               case 131: // auto-detect delle finestre 
                  find_window(&WR,&WC,&WL,&WA);
                  set_pkt(132);
                  break;
               case 133: // attivazione di una finestra 
                  x_min=pkt.data[0];
                  y_min=pkt.data[1];
                  x_max=pkt.data[2];
                  y_max=pkt.data[3];
                  if ((x_min==0) && (y_min==0) && 
                      (x_max==(n_col-1)) && (y_max==(n_lines-1)))
                     full_screen=TRUE;
                  else 
                     full_screen=FALSE;
                  x_min0=x_min;
                  y_min0=y_min;
                  x_max0=x_max;
                  y_max0=y_max;
                  mb_x=pkt.data[4];
                  mb_y=pkt.data[5];
                  get_side_steps();
                  set_mb_position(0,up_v,&ok);
                  video_changed1(start_adrs);
                  set_pkt(136);
                  break;   
               case 137: // parametri setup cursor routing 
                     if (pkt.data[0]==CUR_ROUTE) rx_crpar();
                  break;
               case 152: // gestione Autoread 
                  auto_read=pkt.data[0];
                  if (cur_y!=mb_y) 
                     type_auto_read=1; 
                  else {
                     if ((cur_x>=mb_x) && (cur_x<(mb_x+40)))
                        type_auto_read=0; 
                     else 
                        type_auto_read=1;
                  }
                  if (auto_read) {
                     old_tracking=tracking;
                     tracking=FALSE;
                  } else {
                     tracking=old_tracking;
                     n_cons=0xFF;
                  }
                  break;


            }
            send_ack();
         }
         // Controlla se si sono modificati l'area video o i parametri di controllo 
         if ( control_changed() ) {
            if (n_line_changed || n_col_changed ) {
               set_new_videomode();
               set_pkt(136);
            }
            // Controlla se il cursore  agganciato 
            if (tracking || re_track) track_the_cursor();
            re_track=FALSE;
            set_pkt(134);
            if (video_changed(start_adrs)) set_pkt(136);
         } else {
            // Se l'area video si  modificata preparati a trasmetterla }
            if (video_changed(start_adrs)) set_pkt(135); 
         }
      }
      
      exit_timer:
      pkt_sender_handler();
      cursor_routing();
      usleep(LOOPTIME*1000);
   }
   
   send_packet(151,0,NULL);
   usleep(500000);
   close_program();
   return 0;
}
